"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@typescript-eslint/utils");
const createRule_1 = __importDefault(require("../util/createRule"));
const memoizeComponent = (0, createRule_1.default)({
    name: 'memoize-component',
    meta: {
        docs: {
            description: 'React components should be memoized',
            recommended: 'warn',
        },
        messages: {
            general: 'Memoize this component',
        },
        type: 'problem',
        schema: [
            {
                type: 'object',
                properties: {},
                additionalProperties: false,
            },
        ],
        fixable: 'code',
    },
    defaultOptions: [],
    create(context) {
        const verifyVariableDeclaration = (node) => {
            var _a, _b;
            const declaration = node.declarations[0];
            if (declaration.type === utils_1.AST_NODE_TYPES.VariableDeclarator) {
                const init = declaration.init;
                if (!init) {
                    // Not a component
                    return;
                }
                const name = declaration.id.type === utils_1.AST_NODE_TYPES.Identifier
                    ? declaration.id.name
                    : undefined;
                if (!name ||
                    // Shallow checking for component declaration
                    name.charAt(0) !== name.charAt(0).toUpperCase()) {
                    return;
                }
                const defaultExportNode = (_b = (_a = node.parent.body) === null || _a === void 0 ? void 0 : _a.find) === null || _b === void 0 ? void 0 : _b.call(_a, item => {
                    if (item.type === utils_1.AST_NODE_TYPES.ExportDefaultDeclaration) {
                        const line = context.getSourceCode().getLines()[item.loc.start.line - 1];
                        if (line.includes(name)) {
                            return true;
                        }
                    }
                    return false;
                });
                if (init.type === utils_1.AST_NODE_TYPES.ArrowFunctionExpression) {
                    if (!defaultExportNode) {
                        context.report({
                            node,
                            messageId: 'general',
                            *fix(fixer) {
                                yield fixer.insertTextBefore(init, 'memo(');
                                yield fixer.insertTextAfter(init, ')');
                            },
                        });
                        return;
                    }
                    const isDefaultExportMemoized = (() => {
                        const exportDeclaration = defaultExportNode.declaration;
                        if (exportDeclaration &&
                            exportDeclaration.type ===
                                utils_1.AST_NODE_TYPES.CallExpression) {
                            const callee = exportDeclaration.callee;
                            if (callee.type === utils_1.AST_NODE_TYPES.Identifier &&
                                callee.name === 'memo') {
                                return true;
                            }
                        }
                    })();
                    if (isDefaultExportMemoized) {
                        return;
                    }
                    context.report({
                        node: defaultExportNode,
                        messageId: 'general',
                        *fix(fixer) {
                            yield fixer.insertTextBefore(defaultExportNode.declaration, 'memo(');
                            yield fixer.insertTextAfter(defaultExportNode.declaration, ')');
                        },
                    });
                    return;
                }
                if (init.type === utils_1.AST_NODE_TYPES.CallExpression) {
                    const callee = init.callee;
                    const args = init.arguments;
                    const calleeChain = args
                        .filter(item => item.type === utils_1.AST_NODE_TYPES.CallExpression &&
                        item.callee.type === utils_1.AST_NODE_TYPES.Identifier)
                        .map((item) => item
                        .callee.name);
                    if (callee.type === utils_1.AST_NODE_TYPES.Identifier) {
                        calleeChain.push(callee.name);
                    }
                    if (!calleeChain.includes('memo')) {
                        context.report({
                            node,
                            messageId: 'general',
                            *fix(fixer) {
                                yield fixer.insertTextBefore(init, 'memo(');
                                yield fixer.insertTextAfter(init, ')');
                            },
                        });
                        return;
                    }
                }
            }
        };
        const verifyExportNamedDeclaration = (node) => {
            var _a;
            if (((_a = node.declaration) === null || _a === void 0 ? void 0 : _a.type) === utils_1.AST_NODE_TYPES.VariableDeclaration) {
                verifyVariableDeclaration(node.declaration);
            }
        };
        const findRootNode = (currentNode) => {
            if ([
                utils_1.AST_NODE_TYPES.VariableDeclaration,
                utils_1.AST_NODE_TYPES.ExportNamedDeclaration,
            ].includes(currentNode.type)) {
                return currentNode;
            }
            const parent = currentNode.parent;
            if (!parent) {
                return undefined;
            }
            return findRootNode(parent);
        };
        const processDeclaration = (declaration) => {
            if (declaration.type === utils_1.AST_NODE_TYPES.VariableDeclaration) {
                verifyVariableDeclaration(declaration);
                return;
            }
            if (declaration.type === utils_1.AST_NODE_TYPES.ExportNamedDeclaration) {
                verifyExportNamedDeclaration(declaration);
            }
        };
        return {
            JSXElement(node) {
                var _a;
                if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) !== utils_1.AST_NODE_TYPES.ReturnStatement) {
                    return;
                }
                const root = findRootNode(node);
                if (!root) {
                    return;
                }
                processDeclaration(root);
            },
            JSXFragment(node) {
                var _a;
                if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) !== utils_1.AST_NODE_TYPES.ReturnStatement) {
                    return;
                }
                const root = findRootNode(node);
                if (!root) {
                    return;
                }
                processDeclaration(root);
            },
        };
    },
});
exports.default = memoizeComponent;
